from slicesampling import univariate_sampler
import numpy as np


def map_1d_to_nd(u, v, x0):
    """

    :param u:
    :param v:
    :param x0:
    :return:
    """

    x = u * v + x0

    return x


class SliceSampler(univariate_sampler.UnivariateSliceSampler):
    """ Implementation of the multivariate slice_sampler sampler, based on [1, 2, 3].

    References
    ----------
    [1] Neal, Radford M. (2003). Slice sampling. The Annals of Statistics, 31(3), 705 - 767. URL: https://doi.org/10.1214/aos/1056562461
    [2] Thompson, Madeleine (2011). Slice Sampling with Multivariate Steps. PhD Thesis. URL: https://hdl.handle.net/1807/31955
    [3] Rudolf, Daniel; Ullrich, Mario (2018). Comparison of hit-and-run, slice_sampler sampler and random walk metropolis. Journal of Applied Probability, 55(4), 1186-1202. URL: https://doi.org/10.1017/jpr.2018.78

    Attributes
    ----------
    :param current_sample = the last known sample
    :param current_log_likelihood_value = log_likelihood of the last known sample
    :param dimension = dimension of the sampled space (number of variables/parameters)
    :param expansion_method = method to expand the sampling interval
    :param expansion_limit = maximum number of expansions for a single sample
    :param initial_interval_width = initial width for the sampling interval
    :param count = number of samples taken since initialization
    """

    def __init__(self, *, log_likelihood_function, initial_sample=np.array(0),
                 adaptation_interval=False,
                 direction_method='random',
                 expansion_method='double',
                 expansion_limit=8,
                 initial_interval_width=1):
        """ Create a multivariate slice_sampler sampler that samples in one direction at the time

        Parameters
        ----------
        :param log_likelihood_function: function that evaluates log-likelihood taking a sample (x) as an input
        :param initial_sample: initial sample (x)
        :param adaptation_interval: number of samples between adaptation of the covariance matrix (0 for no updates)
        :param direction_method: method to sample in multivariate space, either
            "univar_random" (sample one random direction according of the covariance matrix for one sample) or
            "univar_eigen" (univariate updates along every eigenvector of the covariance matrix for one sample)
        :param expansion_method: method to expand sampling interval (either 'fixed', 'double', or 'step_out')
        :param expansion_limit: maximum number of expansions of sampling interval for one new sample, only relevant if
        expansion_method equals 'double' or 'step_out'.
        :param initial_interval_width: initial width of sampling interval, only relevant if
        expansion_method equals 'double' or 'step_out'.
        """

        assert isinstance(initial_sample, np.ndarray), \
            "initial_sample must be a NumPy array (numpy.ndarray). "
        assert initial_sample.ndim == 1, \
            "initial_sample must be a 1-dimensional array."
        assert expansion_method in ['fixed', 'double', 'step_out'], \
            "method to choose direction must be either 'random' or 'gibbslike'. "
        assert direction_method in ['random_eigen', 'univar_eigen'], \
            "method to choose direction must be either 'random_eigen' or 'univar_eigen'. "

        super().__init__(log_likelihood_function=log_likelihood_function,
                         initial_sample=initial_sample,
                         expansion_method=expansion_method,
                         expansion_limit=expansion_limit,
                         initial_interval_width=initial_interval_width)

        dimension = initial_sample.shape[0]

        if dimension == 1:
            # override adaptation, only needed in multivariate case
            adaptation_interval = 0
            pass

        self.adaptation_interval = adaptation_interval
        self.count = 0
        self.sample_mean = initial_sample
        self.sample_covariance = np.zeros(dimension)
        self.sample_coordinate_basis = np.eye(dimension)

        self.dimension = dimension
        self.multivariate_log_likelihood_function = log_likelihood_function

        self.direction_method = direction_method

        self._assert_input()

        return

    def _assert_input(self):

        assert isinstance(self.current_sample, np.ndarray), "initial_sample must be a float. "

        return

    def _get_sample_univar(self, v, x0, y0):
        """ Apply univariate slice_sampler sampler to get a sample in the chose direction

        :param v: chosen direction and unit vector
        :param x0: current sample in the multivariate space
        :param y0: log-likelihood at the current sample
        :return: new sample as pair of sample (x1) and log-likelihood (y1)
        """

        self.log_likelihood_function = lambda u: self.multivariate_log_likelihood_function(map_1d_to_nd(u, v, x0))
        u0 = 0
        [u1, y1] = self._get_sample(u0, y0)
        x1 = map_1d_to_nd(u1, v, x0)

        return x1, y1

    def _sample_update(self):
        """ Get a new sample and update associated attributes in the sampler object and execute following actions as
        necessary:
        -   recursive estimation of mean vector and covariance matrix of the samples
        -   adjust coordinate basis for the multivariate space so origin reflects the sample mean and the coordinates
            align with the eigenvectors of the covariance matrix
        """

        x0 = self.current_sample
        y0 = self.current_log_likelihood_value

        if self.direction_method == 'random_eigen':
            v = np.matmul(self.sample_coordinate_basis, np.random.randn(self.dimension))
            [x0, y0] = self._get_sample_univar(v, x0, y0)
            pass
        elif self.direction_method == 'univar_eigen':
            for direction in np.random.permutation(range(self.dimension)):
                v = self.sample_coordinate_basis[:, direction]
                [x0, y0] = self._get_sample_univar(v, x0, y0)
                pass
            pass

        self.current_sample = x0
        self.current_log_likelihood_value = y0
        self.count += 1

        if self.adaptation_interval > 0:
            alpha = 1 / self.count
            dx = x0 - self.sample_mean
            self.sample_covariance = (1 - alpha) * (self.sample_covariance + alpha * np.dot(dx[:, None], dx[:, None].T))
            self.sample_mean = (1 - alpha) * self.sample_mean + alpha * x0
            if np.mod(self.count, self.adaptation_interval) == 0:
                eigenvectors, eigenvalues, eigenvectorsT = np.linalg.svd(self.sample_covariance)
                coordinate_basis = np.matmul(eigenvectors, np.diag(eigenvalues ** (1 / 2)))
                self.sample_coordinate_basis = coordinate_basis
                pass
            pass

        return

    pass


if __name__ == '__main__':
    help(SliceSampler)
    pass



#!/usr/bin/env python


import numpy as np
import warnings


class UnivariateSliceSampler:
    """
    Implementation of the univariate slice_sampler sampler, as described in [1].

    References
    ----------
    [1] Neal, R. M. (2003). Slice sampling. Ann. Statist. 31 (3) 705 - 767. https://doi.org/10.1214/aos/1056562461

    Attributes
    ----------
    :param current_sample = the last known sample
    :param current_log_likelihood_value = log_likelihood of the last known sample
    :param dimension = dimension of the sampled space (number of variables/parameters)
    :param expansion_method = method to expand the sampling interval
    :param expansion_limit = maximum number of expansions for a single sample
    :param initial_interval_width = initial width for the sampling interval
    :param count = number of samples taken since initialization
    """

    def __init__(self, *, log_likelihood_function,
                 initial_sample=0,
                 expansion_method: str = 'double',
                 expansion_limit: int = 8,
                 initial_interval_width: float = 1):
        """
        Create a univariate slice_sampler sampler

        Parameters
        ----------
        :param log_likelihood_function: function that evaluates log-likelihood taking a sample (x) as an input
        :param initial_sample: initial sample (x)
        :param expansion_method: method to expand sampling interval (either 'fixed', 'double', or 'step_out')
        :param expansion_limit: maximum number of expansions of sampling interval for one new sample, only relevant if
        expansion_method equals 'double' or 'step_out'.
        :param initial_interval_width: initial width of sampling interval, only relevant if
        expansion_method equals 'double' or 'step_out'.
        """

        assert expansion_method in ['fixed', 'double', 'step_out'], \
            "method to choose direction must be either 'fixed', 'double', or 'step_out'. "

        self.log_likelihood_function = log_likelihood_function

        if np.isnan(log_likelihood_function(initial_sample)):
            warnings.warn("Initial sample results in NaN results for log_likelihood_function. Sampler may not succeed "
                          "in finding feasible samples.")
            pass

        if np.isinf(log_likelihood_function(initial_sample)):
            warnings.warn("Initial sample results in infinity value for log_likelihood_function. Sampler may not "
                          "succeed in finding feasible samples.")
            pass

        self.current_sample = initial_sample
        self.current_log_likelihood_value = self.log_likelihood_function(initial_sample)
        self.dimension = 1
        self.expansion_method = expansion_method
        self.expansion_limit = expansion_limit
        self.initial_interval_width = initial_interval_width
        self.count = 0

        self._assert_input()

        return

    def _assert_input(self):

        assert isinstance(self.current_sample, float), "initial_sample must be a float. "

        return

    def get_trace(self, number_of_samples: int = 1000):
        """
        Produces a sequence of Markov Chain Monte Carlo samples with the slice_sampler sampler

        :param number_of_samples: number of samples to get from the sampler
        :return: sequence of samples, corresponding log-likelihood values
        """

        sample_sequence = np.matrix(np.zeros([0, self.dimension]))
        log_likelihood_seq = []
        sample_index = 0
        while sample_index < number_of_samples:
            self._sample_update()
            sample = self.current_sample
            log_likelihood_value = self.current_log_likelihood_value
            if np.isinf(log_likelihood_value) or np.isnan(log_likelihood_value):
                # do not add sample to sequence if infeasible
                pass
            else:
                sample_sequence = np.append(sample_sequence, np.matrix(sample), axis=0)
                log_likelihood_seq = np.append(log_likelihood_seq, log_likelihood_value)
                sample_index += 1
                pass

        return sample_sequence, log_likelihood_seq

    def _sample_update(self):
        """ Get a new sample and update associated attributes in the sampler object """

        x0 = self.current_sample
        y0 = self.current_log_likelihood_value

        [x1, y1] = self._get_sample(x0, y0)

        self.current_sample = x1
        self.current_log_likelihood_value = y1
        self.count += 1
        return

    def _get_sample(self, x0, y0):
        """ Get a new Monte Carlo sample starting from a given sample (x0) and corresponding log-likelihood (y0)

        :param x0: sample to start with
        :param y0: auxiliary variable (log-likelihood) to start with
        :return: [x1, y1]; pair of new sample (x1) and corresponding log-likelihood (y1)
        """

        # define auxiliary variable range as [0, y]
        alpha = np.random.rand(1)
        y = np.log(alpha) + y0

        # apply procedure to define new sampling interval
        if self.expansion_method == 'step_out':
            [bound_left, bound_right] = self._step_out(x0, y)
            pass
        elif self.expansion_method == 'double':
            [bound_left, bound_right] = self._double(x0, y)
            pass
        elif self.expansion_method == 'fixed':
            u = np.random.rand(1)
            bound_left = x0 - self.initial_interval_width * u
            bound_right = bound_left + self.initial_interval_width
            pass

        # sample from sampling interval defined above
        accept_sample = False
        while not accept_sample:

            # get proposal
            u = np.random.rand(1)
            x1 = bound_left + u * (bound_right - bound_left)
            y1 = self.log_likelihood_function(x1)

            # evaluate proposal
            if y < y1 and self._accept(x0, x1, y, bound_left, bound_right):
                accept_sample = True
                pass

            # update sampling interval if proposal was rejected
            if not accept_sample:
                if x1 < x0:
                    bound_left = x1
                    pass
                else:
                    bound_right = x1
                    pass
                pass

            pass

        return x1, y1

    def _accept(self, x0, x1, y, bound_left, bound_right):
        """ Evaluate whether proposed sample satisfies detailed balance

        :param x0: sample of origin
        :param x1: newly proposed sample
        :param y: auxiliary variable (log-likelihood reference)
        :param bound_left: left bound of sampling interval
        :param bound_right: right bound of sampling interval
        :return: accept/reject decision
        :rtype: boolean
        """

        # accept sample by default
        accept_sample = True

        # detailed balance check when using the doubling interval expansion method. See original paper cited in class
        # docstring for detailed information
        if self.expansion_method == 'double':
            bound_left_star = bound_left
            bound_right_star = bound_right
            y_left_star = np.nan
            y_right_star = np.nan
            decision = False
            while bound_right_star - bound_left_star > 1.1 * self.initial_interval_width:
                center = (bound_left_star + bound_right_star) / 2
                if (x0 < center <= x1) or (x0 >= center > x1):
                    decision = True
                    pass
                if x1 < center:
                    bound_right_star = center
                    y_right_star = self.log_likelihood_function(bound_right_star)
                    pass
                else:
                    bound_left_star = center
                    y_left_star = self.log_likelihood_function(bound_left_star)
                    pass
                if decision:
                    if np.isnan(y_left_star):
                        y_left_star = self.log_likelihood_function(bound_left_star)
                        pass
                    if np.isnan(y_right_star):
                        y_right_star = self.log_likelihood_function(bound_right_star)
                        pass
                    if y >= y_left_star and y >= y_right_star:
                        accept_sample = False
                        pass
                    pass

                pass
            pass

        return accept_sample

    def _double(self, x0, y):
        """ Apply doubling procedure to obtain new sampling interval

        :param x0: current sample
        :param y: auxiliary variable (reference log-likelihood)
        :return: sampling interval as a pair of scalars: bound_left, bound_right
        """

        # define initial interval by picking a random number from the unit interval and defining this point as the
        # center of the initial sampling interval
        u = np.random.rand(1)
        bound_left = x0 - self.initial_interval_width * u
        bound_right = bound_left + self.initial_interval_width

        # evaluate log-likelihood at the interval bounds
        y_left = self.log_likelihood_function(bound_left)
        y_right = self.log_likelihood_function(bound_right)

        # continue to expand the sampling interval until (a) the interval bound log-likelihood remain above the
        # reference log-likelihood or (b) the maximum number of expansions has been met
        expansion_budget = self.expansion_limit
        while expansion_budget > 0 and (y < y_left or y < y_right):
            # determine to expand to the left (0) or to the right (1)
            v = np.random.rand(1) * 2
            direction = v > 1

            # double the size of the interval
            if direction:
                bound_right = bound_right + (bound_right - bound_left)
                y_right = self.log_likelihood_function(bound_right)
                pass
            else:
                bound_left = bound_left - (bound_right - bound_left)
                y_left = self.log_likelihood_function(bound_left)
                pass

            # update number of expansions available
            expansion_budget -= 1
            pass

        return bound_left, bound_right

    def _step_out(self, x0, y):
        """ Apply step-out procedure to obtain new sampling interval

        :param x0: current sample
        :param y: auxiliary variable (reference log-likelihood)
        :return: sampling interval as a pair of scalars: bound_left, bound_right
        """

        # define initial interval by picking a random number from the unit interval and defining this point as the
        # center of the initial sampling interval
        u = np.random.rand(1)
        bound_left = x0 - self.initial_interval_width * u
        bound_right = bound_left + self.initial_interval_width

        # determine number of times one can step out to the left and to the right
        v = np.random.rand(1)
        expansion_budget_left = np.floor(self.expansion_limit * v)
        expansion_budget_right = (self.expansion_limit - 1) - expansion_budget_left
        while expansion_budget_left > 0 and y < self.log_likelihood_function(bound_left):
            bound_left = bound_left - self.initial_interval_width
            expansion_budget_left = expansion_budget_left - 1
            pass

        while expansion_budget_right > 0 and y < self.log_likelihood_function(bound_right):
            bound_right = bound_right + self.initial_interval_width
            expansion_budget_right = expansion_budget_right - 1
            pass

        return bound_left, bound_right

    pass


if __name__ == '__main__':
    help(UnivariateSliceSampler)
    pass



from setuptools import setup

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

exec(open('slicesampling/version.py').read())

setup(name='slicesampling',
      version=__version__,
      description='Set of Markov chain Monte Carlo (MCMC) sampling methods based on slice_sampler sampling',
      url='https://code.ornl.gov/2kv/slicesampler',
      author='Kris Villez',
      author_email='villezk@ornl.gov',
      install_requires=['numpy', 'scipy'],
      license='MIT',
      packages=['slicesampling'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False)